﻿using System;
using System.IO;

namespace msxdisk
{
    class Cassette
    {
        public const int frameSize = 0x40;
        byte[] header = new byte[] { 0x1f, 0xa6, 0xde, 0xba, 0xcc, 0x13, 0x7d, 0x74 };

        static byte[][][] patches = new byte[][][]
            {
                //new byte [][] //OUT (0A8h),A - 1 ERROR IN 64KB, 1/2 ERROR IN 32KB
                //{
                //    new byte[] { 0xd3, 0xa8 },
                //    new byte[] { 0x00, 0x00 }
                //},
                new byte [][] //PACMAN, AFTERBURNER
                {
                    new byte[] { 0xdb, 0xa8, 0xf5, 0xe6, 0xfc, 0xd3, 0xa8 },
                    new byte[] { 0xdb, 0xa8, 0xf5, 0xe6, 0xfc, 0x00, 0x00 }
                },
                //new byte [][] //SOCCER
                //{
                //    new byte[] { 0xDB, 0xA8, 0x4F, 0xE6, 0x3F, 0xD3, 0xA8 },
                //    new byte[] { 0xDB, 0xA8, 0x4F, 0xE6, 0x3F, 0x00, 0x00 }
                //},
                //new byte [][] //LD DE,0D2FEh; JP NZ,0D1FEh; Passette loader - Disable Checksum 
                //{
                //    new byte[] { 0x11, 0xfe, 0xd2, 0xc2, 0xfe, 0xd1 },
                //    new byte[] { 0x11, 0xfe, 0xd2, 0x00, 0x00, 0x00 }
                //},
                //new byte [][] //RET Z; LD DE,0D2FEh; JP 0D1FEh; Passette loader - Disable Checksum
                //{
                //    new byte[] { 0xc8, 0x11, 0xfe, 0xd2, 0xc3, 0xfe, 0xd1 },
                //    new byte[] { 0xc9, 0x11, 0xfe, 0xd2, 0xc3, 0xfe, 0xd1 }
                //},
            };

        byte[] cas;
        int pos = 0;

        FileStream fsCAS;
        FileStream fsOut;

        object lockOut = new object();

        public Cassette(string path)
        {
            try
            {
                fsCAS = new FileStream(path, FileMode.Open, FileAccess.ReadWrite, FileShare.None);
            }
            catch (UnauthorizedAccessException)
            {
                fsCAS = new FileStream(path, FileMode.Open, FileAccess.Read, FileShare.None);
            }

            load();
        }


        void load()
        {
            cas = new byte[fsCAS.Length];

            fsCAS.Seek(0, SeekOrigin.Begin);
            fsCAS.Read(cas, 0, cas.Length);

            //Apply Patches

            for (int i = 0; i < cas.Length; i++) //byte
            {
                for (int j = 0; j < patches.Length; j++) //patch
                {
                    if (i + patches[j][0].Length > cas.Length)
                        goto next_j;

                    for (int k = 0; k < patches[j][0].Length; k++) //byte in patch
                    {
                        if (cas[i + k] != patches[j][0][k])
                            goto next_j;
                    }

                    for (int k = 0; k < patches[j][0].Length; k++) //byte in patch
                    {
                        cas[i + k] = patches[j][1][k];
                    }

                next_j:;
                }
            }
        }

        bool advance()
        {
            if (pos >= cas.Length)
                return false;
            pos += 8;
            return true;
        }

        bool is_header()
        {
            if (pos + 7 >= cas.Length)
                return false;
            for (int i = 0; i < 8; i++)
                if (cas[pos + i] != header[i])
                    return false;
            return true;
        }

        public bool tapion()
        {
            for (; ;)
            {
                if (is_header())
                {
                    advance();
                    return true;
                }
                if (!advance())
                    return false;
            }
        }

        public bool tapin(byte[] frame)
        {
            int start = pos;
            while (pos - start < frameSize)
            {
                if (is_header())
                    break;
                if (!advance())
                    break;
            }
            int bytes = (pos > cas.Length ? cas.Length : pos) - start;
            if (bytes == 0) return false;
            Array.Copy(cas, start, frame, 0, bytes);
            for (int i = bytes; i < frameSize; i++)
                frame[i] = 0x1A;
            return true;
        }

        public bool tapiof()
        {
            return true;
        }

        public bool tapoon()
        {
            try
            {
                if (!fsCAS.CanWrite) return false;

                if (fsOut != null) tapoof();

                fsOut = new FileStream("TAPOUT.TMP", FileMode.Create, FileAccess.ReadWrite);
                fsOut.Write(header, 0, header.Length);
            }
            catch
            {
                return false;
            }

            return true;
        }

        public bool tapout_ok()
        {
            return fsOut != null;
        }

        public bool tapout(byte[] frame)
        {
            try
            {
                fsOut.Write(frame, 0, frameSize);
            }
            catch
            {
                return false;
            }

            return true;
        }

        public bool tapoof()
        {
            try
            {
                byte[] tmp = new byte[fsOut.Length];

                fsOut.Seek(0, SeekOrigin.Begin);
                fsOut.Read(tmp, 0, tmp.Length);

                lock (lockOut)
                {
                    fsOut.Close();
                    File.Delete("OUT.TMP");
                    fsOut = null;
                }

                fsCAS.Seek(pos, SeekOrigin.Begin);
                fsCAS.Write(tmp, 0, tmp.Length);

                pos += tmp.Length;

                load();
            }
            catch
            {
                return false;
            }

            return true;
        }

        public void rewind()
        {
            pos = 0;
        }

        public void get_pos(out int pos, out int len)
        {
            lock (lockOut)
            {
                if (fsOut != null)
                {
                    pos = this.pos + (int)fsOut.Length;
                    len = (int)Math.Max(fsCAS.Length, pos);
                }
                else
                {
                    pos = this.pos;
                    len = cas.Length;
                }
            }
        }

        public void close()
        {
            fsCAS.Close();
        }
    }
}
