﻿using System;
using System.ComponentModel;
using System.Windows.Forms;
using System.Text;
using System.IO;
using System.Collections.Generic;
using System.Drawing;

namespace msxdisk
{
    public partial class ChestForm : Form
    {
        public class Item
        {
            public enum Status
            {
                New,
                Existing,
                Deleted
            }

            public string type;
            public string file;
            public string title;
            public string year;
            public string developer;
            public string command;
            public string notes;

            public Status status = Status.Existing;
        }

        public ChestForm()
        {
            InitializeComponent();
        }

        Control searchControl;

        string cmbSearchType_Text;
        string cmbSearchTitle_Text;
        string cmbSearchYear_Text;
        string cmbSearchDeveloper_Text;
        bool apply = false;
        bool sync = false;

        public Item SelectedItem;

        private void ChestForm_Shown(object sender, EventArgs e)
        {
            searchControl = cmbSearchType;

            cmbSearchType_Text = cmbSearchType.Text;
            cmbSearchTitle_Text = cmbSearchTitle.Text;
            cmbSearchYear_Text = cmbSearchYear.Text;
            cmbSearchDeveloper_Text = cmbSearchDeveloper.Text;

            worker.RunWorkerAsync();
        }

        private void ChestForm_KeyDown(object sender, KeyEventArgs e)
        {
            if (this.ActiveControl.Parent == pnlSearch)
            {
                switch (e.KeyCode)
                {
                    case Keys.Enter:
                        btnSearchApply.PerformClick();
                        e.SuppressKeyPress = true;
                        break;
                    case Keys.Escape:
                        btnSearchCancel.PerformClick();
                        e.SuppressKeyPress = true;
                        break;
                    case Keys.F5:
                        btnSearchSync.PerformClick();
                        e.SuppressKeyPress = true;
                        break;
                }
            }
            else if (this.ActiveControl == lstChest)
            {
                switch (e.KeyCode)
                {
                    case Keys.Enter:
                        if (btnMount.Enabled)
                        {
                            btnMount.PerformClick();
                            e.SuppressKeyPress = true;
                        }
                        break;
                    case Keys.Escape:
                        if (this.ControlBox)
                        {
                            this.Close();
                            e.SuppressKeyPress = true;
                        }
                        break;
                }
            }
            else if (this.ActiveControl.Parent == pnlDetails)
            {
                switch (e.KeyCode)
                {
                    case Keys.Enter:
                        if (this.ActiveControl != txtNotes && btnSave.Enabled)
                        {
                            btnSave.PerformClick();
                            e.SuppressKeyPress = true;
                        }
                        break;
                    case Keys.Escape:
                        if (btnCancel.Enabled)
                        {
                            btnCancel.PerformClick();
                            e.SuppressKeyPress = true;
                        }
                        break;
                }
            }
        }

        private void ChestForm_FormClosed(object sender, FormClosedEventArgs e)
        {
            Clear();
        }

        private void btnSearchApply_Click(object sender, EventArgs e)
        {
            searchControl = this.ActiveControl;

            cmbSearchType_Text = cmbSearchType.Text;
            cmbSearchTitle_Text = cmbSearchTitle.Text;
            cmbSearchYear_Text = cmbSearchYear.Text;
            cmbSearchDeveloper_Text = cmbSearchDeveloper.Text;
            apply = true;
            sync = false;

            worker.RunWorkerAsync();
        }

        private void btnSearchCancel_Click(object sender, EventArgs e)
        {
            searchControl = this.ActiveControl;

            cmbSearchType.Text = "";
            cmbSearchTitle.Text = "";
            cmbSearchYear.Text = "";
            cmbSearchDeveloper.Text = "";

            cmbSearchType_Text = cmbSearchType.Text;
            cmbSearchTitle_Text = cmbSearchTitle.Text;
            cmbSearchYear_Text = cmbSearchYear.Text;
            cmbSearchDeveloper_Text = cmbSearchDeveloper.Text;
            apply = false;
            sync = false;

            worker.RunWorkerAsync();
        }

        private void btnSearchSync_Click(object sender, EventArgs e)
        {
            searchControl = this.ActiveControl;

            cmbSearchType.Text = "";
            cmbSearchTitle.Text = "";
            cmbSearchYear.Text = "";
            cmbSearchDeveloper.Text = "";

            cmbSearchType_Text = cmbSearchType.Text;
            cmbSearchTitle_Text = cmbSearchTitle.Text;
            cmbSearchYear_Text = cmbSearchYear.Text;
            cmbSearchDeveloper_Text = cmbSearchDeveloper.Text;
            apply = true;
            sync = true;

            worker.RunWorkerAsync();
        }

        private void lstChest_SelectedIndexChanged(object sender, EventArgs e)
        {
            if (lstChest.SelectedItems.Count == 1)
            {
                ListViewItem lviChest = lstChest.SelectedItems[0];
                Item item = (Item)lviChest.Tag;

                txtType.Text = item.type;
                txtFile.Text = item.file;
                txtTitle.Text = item.title;
                txtYear.Text = item.year;
                txtDeveloper.Text = item.developer;
                txtCommand.Text = item.command;
                txtNotes.Text = item.notes;

                btnMount.Enabled = true;
                btnEdit.Enabled = true;
                btnDelete.Enabled = true;
            }
            else
            {
                txtType.Text = "";
                txtFile.Text = "";
                txtTitle.Text = "";
                txtYear.Text = "";
                txtDeveloper.Text = "";
                txtCommand.Text = "";
                txtNotes.Text = "";

                btnMount.Enabled = false;
                btnEdit.Enabled = false;
                btnDelete.Enabled = false;
            }
        }

        private void btnMount_Click(object sender, EventArgs e)
        {
            SelectedItem = (Item)lstChest.SelectedItems[0].Tag;
        }

        private void btnNew_Click(object sender, EventArgs e)
        {
            ofd.FileName = "";
            if (ofd.ShowDialog(this) == DialogResult.OK)
            {
                string chestDir = Path.GetFullPath("CHEST");
                string file;
                string extension;

                try
                {
                    if (Path.GetDirectoryName(ofd.FileName) == chestDir)
                    {
                        throw new Exception("File is already in CHEST directory.");
                    }
                    else
                    {
                        string fileName = Path.GetFileNameWithoutExtension(ofd.FileName);
                        extension = Path.GetExtension(ofd.FileName);
                        string dest;

                        for (int i = 1; ; i++)
                        {
                            dest = Path.Combine(chestDir, fileName + (i == 1 ? "" : " (" + i.ToString() + ")") + extension);
                            if (!File.Exists(dest)) break;
                        }
                        File.Copy(ofd.FileName, dest);

                        file = Path.GetFileName(dest);
                    }
                }
                catch (Exception ex)
                {
                    Error(ex.Message);
                    return;
                }

                ListViewItem lviChest = new ListViewItem(new string[] { "", "", "", "" });

                Item item = new Item();
                item.type = extension.StartsWith(".") ? extension.Substring(1).ToUpper() : "";
                item.file = "";
                item.title = "";
                item.year = "";
                item.developer = "";
                item.command = "";
                item.notes = "";
                item.status = Item.Status.New;

                lviChest.Tag = item;

                this.lstChest.Items.Add(lviChest);

                foreach (ListViewItem lviSelected in lstChest.SelectedItems)
                    lviSelected.Selected = false;
                lviChest.Selected = true;

                lviChest.EnsureVisible();

                txtFile.Text = file;

                btnEdit.PerformClick();
            }
        }

        private void btnEdit_Click(object sender, EventArgs e)
        {
            this.ControlBox = false;
            pnlSearch.Enabled = false;
            lstChest.Enabled = false;
            btnNew.Enabled = false;
            btnEdit.Enabled = false;
            btnDelete.Enabled = false;

            txtTitle.ReadOnly = false;
            txtYear.ReadOnly = false;
            txtDeveloper.ReadOnly = false;
            txtCommand.ReadOnly = false;
            txtNotes.ReadOnly = false;

            btnSave.Enabled = true;
            btnCancel.Enabled = true;

            txtYear.Focus();
        }

        private void btnDelete_Click(object sender, EventArgs e)
        {
            ListViewItem lviChest = lstChest.SelectedItems[0];
            Item item = (Item)lviChest.Tag;

            if (MessageBox.Show(this, "You are about to delete:\r\n" + item.title, "Confirm", MessageBoxButtons.OKCancel, MessageBoxIcon.Warning, MessageBoxDefaultButton.Button2) == DialogResult.OK)
            {
                try
                {
                    File.Delete(Path.Combine("CHEST", item.file));

                    string inCSV = FileHelper.ReadFile(Path.Combine("CHEST", "_.CSV"));
                    StringBuilder outCSV = new StringBuilder(Convert.ToInt32(inCSV.Length * 1.1));
                    int pos = 0;
                    bool eof = false;
                    string[] csv = FileHelper.GetCsvValues(inCSV, ref pos, out eof);

                    outCSV.Append(FileHelper.ToCsvValue(csv[0]));
                    outCSV.Append(',');
                    outCSV.Append(FileHelper.ToCsvValue(csv[1]));
                    outCSV.Append(',');
                    outCSV.Append(FileHelper.ToCsvValue(csv[2]));
                    outCSV.Append(',');
                    outCSV.Append(FileHelper.ToCsvValue(csv[3]));
                    outCSV.Append(',');
                    outCSV.Append(FileHelper.ToCsvValue(csv[4]));
                    outCSV.Append(',');
                    outCSV.Append(FileHelper.ToCsvValue(csv[5]));
                    outCSV.Append(',');
                    outCSV.Append(FileHelper.ToCsvValue(csv[6]));
                    outCSV.Append("\r\n");

                    while (!eof)
                    {
                        csv = FileHelper.GetCsvValues(inCSV, ref pos, out eof);

                        if (csv[1] != item.file)
                        {
                            outCSV.Append(FileHelper.ToCsvValue(csv[0]));
                            outCSV.Append(',');
                            outCSV.Append(FileHelper.ToCsvValue(csv[1]));
                            outCSV.Append(',');
                            outCSV.Append(FileHelper.ToCsvValue(csv[2]));
                            outCSV.Append(',');
                            outCSV.Append(FileHelper.ToCsvValue(csv[3]));
                            outCSV.Append(',');
                            outCSV.Append(FileHelper.ToCsvValue(csv[4]));
                            outCSV.Append(',');
                            outCSV.Append(FileHelper.ToCsvValue(csv[5]));
                            outCSV.Append(',');
                            outCSV.Append(FileHelper.ToCsvValue(csv[6]));
                            outCSV.Append("\r\n");
                        }
                    }
                    FileHelper.WriteFile(Path.Combine("CHEST", "_.CSV"), outCSV.ToString());
                }
                catch (Exception ex)
                {
                    Error(ex.Message);
                    return;
                }

                lviChest.Remove();
            }
        }

        private void btnSave_Click(object sender, EventArgs e)
        {
            ListViewItem lviChest = lstChest.SelectedItems[0];
            Item item = (Item)lviChest.Tag;

            try
            {
                string inCSV = FileHelper.ReadFile(Path.Combine("CHEST", "_.CSV"));
                StringBuilder outCSV = new StringBuilder(Convert.ToInt32(inCSV.Length * 1.1));
                int pos = 0;
                bool eof = false;
                string[] csv = FileHelper.GetCsvValues(inCSV, ref pos, out eof);

                outCSV.Append(FileHelper.ToCsvValue(csv[0]));
                outCSV.Append(',');
                outCSV.Append(FileHelper.ToCsvValue(csv[1]));
                outCSV.Append(',');
                outCSV.Append(FileHelper.ToCsvValue(csv[2]));
                outCSV.Append(',');
                outCSV.Append(FileHelper.ToCsvValue(csv[3]));
                outCSV.Append(',');
                outCSV.Append(FileHelper.ToCsvValue(csv[4]));
                outCSV.Append(',');
                outCSV.Append(FileHelper.ToCsvValue(csv[5]));
                outCSV.Append(',');
                outCSV.Append(FileHelper.ToCsvValue(csv[6]));
                outCSV.Append("\r\n");

                bool found = false;

                while (!eof)
                {
                    csv = FileHelper.GetCsvValues(inCSV, ref pos, out eof);

                    if (csv[1] == item.file)
                    {
                        found = true;

                        outCSV.Append(FileHelper.ToCsvValue(txtType.Text));
                        outCSV.Append(',');
                        outCSV.Append(FileHelper.ToCsvValue(txtFile.Text));
                        outCSV.Append(',');
                        outCSV.Append(FileHelper.ToCsvValue(txtTitle.Text));
                        outCSV.Append(',');
                        outCSV.Append(FileHelper.ToCsvValue(txtYear.Text));
                        outCSV.Append(',');
                        outCSV.Append(FileHelper.ToCsvValue(txtDeveloper.Text));
                        outCSV.Append(',');
                        outCSV.Append(FileHelper.ToCsvValue(txtCommand.Text));
                        outCSV.Append(',');
                        outCSV.Append(FileHelper.ToCsvValue(txtNotes.Text));
                    }
                    else
                    {
                        outCSV.Append(FileHelper.ToCsvValue(csv[0]));
                        outCSV.Append(',');
                        outCSV.Append(FileHelper.ToCsvValue(csv[1]));
                        outCSV.Append(',');
                        outCSV.Append(FileHelper.ToCsvValue(csv[2]));
                        outCSV.Append(',');
                        outCSV.Append(FileHelper.ToCsvValue(csv[3]));
                        outCSV.Append(',');
                        outCSV.Append(FileHelper.ToCsvValue(csv[4]));
                        outCSV.Append(',');
                        outCSV.Append(FileHelper.ToCsvValue(csv[5]));
                        outCSV.Append(',');
                        outCSV.Append(FileHelper.ToCsvValue(csv[6]));
                    }
                    outCSV.Append("\r\n");
                }

                if (!found)
                {
                    outCSV.Append(FileHelper.ToCsvValue(txtType.Text));
                    outCSV.Append(',');
                    outCSV.Append(FileHelper.ToCsvValue(txtFile.Text));
                    outCSV.Append(',');
                    outCSV.Append(FileHelper.ToCsvValue(txtTitle.Text));
                    outCSV.Append(',');
                    outCSV.Append(FileHelper.ToCsvValue(txtYear.Text));
                    outCSV.Append(',');
                    outCSV.Append(FileHelper.ToCsvValue(txtDeveloper.Text));
                    outCSV.Append(',');
                    outCSV.Append(FileHelper.ToCsvValue(txtCommand.Text));
                    outCSV.Append(',');
                    outCSV.Append(FileHelper.ToCsvValue(txtNotes.Text));
                    outCSV.Append("\r\n");
                }

                FileHelper.WriteFile(Path.Combine("CHEST", "_.CSV"), outCSV.ToString());
            }
            catch (Exception ex)
            {
                Error(ex.Message);
                return;
            }

            lviChest.SubItems[0].Text = txtTitle.Text;
            lviChest.SubItems[1].Text = txtType.Text;
            lviChest.SubItems[2].Text = txtYear.Text;
            lviChest.SubItems[3].Text = txtDeveloper.Text;

            item.type = txtType.Text;
            item.file = txtFile.Text;
            item.title = txtTitle.Text;
            item.year = txtYear.Text;
            item.developer = txtDeveloper.Text;
            item.command = txtCommand.Text;
            item.notes = txtNotes.Text;

            if (item.status == Item.Status.New)
            {
                item.status = Item.Status.Existing;
                lviChest.BackColor = SystemColors.Window;
            }

            this.ControlBox = true;
            pnlSearch.Enabled = true;
            lstChest.Enabled = true;
            btnNew.Enabled = true;
            btnEdit.Enabled = true;
            btnDelete.Enabled = true;

            txtFile.ReadOnly = true;
            txtTitle.ReadOnly = true;
            txtYear.ReadOnly = true;
            txtDeveloper.ReadOnly = true;
            txtCommand.ReadOnly = true;
            txtNotes.ReadOnly = true;

            btnSave.Enabled = false;
            btnCancel.Enabled = false;

            lstChest.Focus();
        }

        private void btnCancel_Click(object sender, EventArgs e)
        {
            ListViewItem lviChest = lstChest.SelectedItems[0];
            Item item = (Item)lviChest.Tag;
            if (item.file == "")
            {
                try { File.Delete(Path.Combine("CHEST", txtFile.Text)); }
                catch (Exception ex) { Error(ex.Message); }

                lviChest.Remove();

                if (lstChest.Items.Count > 0)
                    lstChest.Items[0].Selected = true;
            }

            this.ControlBox = true;
            pnlSearch.Enabled = true;
            lstChest.Enabled = true;
            btnNew.Enabled = true;

            //btnEdit.Enabled = true;
            //btnDelete.Enabled = true;
            lstChest_SelectedIndexChanged(lstChest, new EventArgs());

            txtFile.ReadOnly = true;
            txtTitle.ReadOnly = true;
            txtYear.ReadOnly = true;
            txtDeveloper.ReadOnly = true;
            txtCommand.ReadOnly = true;
            txtNotes.ReadOnly = true;

            btnSave.Enabled = false;
            btnCancel.Enabled = false;

            lstChest.Focus();
        }

        private void worker_DoWork(object sender, DoWorkEventArgs e)
        {
            SetEnabled(false);

            Clear();

            try
            {
                string chestCSV = FileHelper.ReadFile(Path.Combine("CHEST", "_.CSV"));
                int pos = 0;
                bool eof = false;

                Dictionary<string, bool> files = new Dictionary<string, bool>();
                foreach (string file in Directory.GetFiles("CHEST"))
                {
                    string fileName = Path.GetFileName(file);
                    string extension = Path.GetExtension(file).ToUpper();
                    if (extension == ".ROM" | extension == ".DSK" | extension == ".CAS")
                        files[fileName] = false;
                }

                FileHelper.GetCsvValues(chestCSV, ref pos, out eof);
                
                while (!eof)
                {
                    string[] csv = FileHelper.GetCsvValues(chestCSV, ref pos, out eof);

                    string type = csv[0];
                    string file = csv[1];
                    string title = csv[2];
                    string year = csv[3];
                    string developer = csv[4];
                    string command = csv[5];
                    string notes = csv[6];

                    if ((cmbSearchType_Text == "" || type.IndexOf(cmbSearchType_Text, StringComparison.OrdinalIgnoreCase) >= 0) &&
                        (cmbSearchTitle_Text == "" || title.IndexOf(cmbSearchTitle_Text, StringComparison.OrdinalIgnoreCase) >= 0) &&
                        (cmbSearchYear_Text == "" || year.IndexOf(cmbSearchYear_Text, StringComparison.OrdinalIgnoreCase) >= 0) &&
                        (cmbSearchDeveloper_Text == "" || developer.IndexOf(cmbSearchDeveloper_Text, StringComparison.OrdinalIgnoreCase) >= 0))
                    {
                        if (files.ContainsKey(file))
                        {
                            files[file] = true;
                            if (!sync)
                                AddItem(type, file, title, year, developer, command, notes, Item.Status.Existing, apply, sync);
                        }
                        else
                        {
                            AddItem(type, file, title, year, developer, command, notes, Item.Status.Deleted, apply, sync);
                        }
                    }
                }

                if (sync)
                    foreach (KeyValuePair<string, bool> file in files)
                        if (!file.Value)
                        {
                            string type = "";
                            string extension = Path.GetExtension(file.Key);
                            if (extension.StartsWith(".")) type = extension.Substring(1).ToUpper();
                            AddItem(type, file.Key, file.Key, "", "", "", "", Item.Status.New, apply, sync);
                        }
            }
            catch (Exception ex)
            {
                Error(ex.Message);
                //return;
            }

            SetEnabled(true);
        }

        delegate void SetEnabledCallback(bool enabled);

        private void SetEnabled(bool enabled)
        {
            if (this.InvokeRequired)
            {
                SetEnabledCallback d = new SetEnabledCallback(SetEnabled);
                this.Invoke(d, new object[] { enabled });
            }
            else
            {
                if (enabled)
                {
                    cmbSearchType.AutoCompleteMode = AutoCompleteMode.Append;
                    cmbSearchTitle.AutoCompleteMode = AutoCompleteMode.Append;
                    cmbSearchYear.AutoCompleteMode = AutoCompleteMode.Append;
                    cmbSearchDeveloper.AutoCompleteMode = AutoCompleteMode.Append;
                    lstChest.Sorting = SortOrder.Ascending;
                    lstChest.Sorting = SortOrder.None;
                    lstChest.EndUpdate();
                }
                else
                {
                    cmbSearchType.AutoCompleteMode = AutoCompleteMode.None;
                    cmbSearchTitle.AutoCompleteMode = AutoCompleteMode.None;
                    cmbSearchYear.AutoCompleteMode = AutoCompleteMode.None;
                    cmbSearchDeveloper.AutoCompleteMode = AutoCompleteMode.None;
                    lstChest.BeginUpdate();
                }

                if (!enabled)
                    txtDummy.Focus();

                this.ControlBox = enabled;
                pnlSearch.Enabled = enabled;
                lstChest.Enabled = enabled;
                btnNew.Enabled = enabled;

                if (enabled)
                    if (apply && lstChest.Items.Count > 0)
                    {
                        lstChest.Focus();
                        lstChest.Items[0].Selected = true;
                    }
                    else
                    {
                        searchControl.Focus();
                    }

                this.UseWaitCursor = !enabled;
            }
        }

        delegate void ClearCallback();

        private void Clear()
        {
            if (this.InvokeRequired)
            {
                ClearCallback d = new ClearCallback(Clear);
                this.Invoke(d, new object[] { });
            }
            else
            {
                lstChest.Items.Clear();
                cmbSearchType.Items.Clear();
                cmbSearchTitle.Items.Clear();
                cmbSearchYear.Items.Clear();
                cmbSearchDeveloper.Items.Clear();

                lstChest_SelectedIndexChanged(lstChest, new EventArgs());
            }
        }

        delegate void AddItemCallback(string type, string file, string title, string year, string developer, string command, string notes, Item.Status status, bool apply, bool sync);

        private void AddItem(string type, string file, string title, string year, string developer, string command, string notes, Item.Status status, bool apply, bool sync)
        {
            if (this.InvokeRequired)
            {
                AddItemCallback d = new AddItemCallback(AddItem);
                this.Invoke(d, new object[] { type, file, title, year, developer, command, notes, status, apply, sync });
            }
            else
            {
                if (!sync)
                {
                    if (type != "" && !cmbSearchType.Items.Contains(type)) cmbSearchType.Items.Add(type);
                    if (title != "" && !cmbSearchTitle.Items.Contains(title)) cmbSearchTitle.Items.Add(title);
                    if (year != "" && !cmbSearchYear.Items.Contains(year)) cmbSearchYear.Items.Add(year);
                    if (developer != "" && !cmbSearchDeveloper.Items.Contains(developer)) cmbSearchDeveloper.Items.Add(developer);
                }

                if (apply)
                {
                    ListViewItem lviChest = new ListViewItem(new string[] { title, type, year, developer });

                    lviChest.BackColor =
                        status == Item.Status.Existing ? SystemColors.Window :
                        status == Item.Status.New ? Color.LightGreen :
                        status == Item.Status.Deleted ? Color.Pink : Color.White;

                    Item item = new Item();
                    item.type = type;
                    item.file = file;
                    item.title = title;
                    item.year = year;
                    item.developer = developer;
                    item.command = command;
                    item.notes = notes;
                    item.status = status;

                    lviChest.Tag = item;

                    this.lstChest.Items.Add(lviChest);
                }
            }
        }

        delegate void ErrorCallback(string message);

        private void Error(string message)
        {
            if (this.InvokeRequired)
            {
                ErrorCallback d = new ErrorCallback(Error);
                this.Invoke(d, new object[] { message });
            }
            else
            {
                MessageBox.Show(this, message, "Error", MessageBoxButtons.OK, MessageBoxIcon.Error);
            }
        }
    }
}
