﻿using System;
using System.Drawing;
using System.Windows.Forms;
using System.Threading;

using System.IO;

namespace msxdisk
{
    public partial class DiskDrive : UserControl
    {
        public DiskDrive()
        {
            InitializeComponent();
        }

        object lockDisk = new object();

        FileStream fsDisk;
        bool dskchg = true;

        int status = 0;

        public void LockDisk()
        {
            Monitor.Enter(lockDisk);
        }

        public void UnlockDisk()
        {
            Monitor.Exit(lockDisk);
        }

        public void SetStatus(int status)
        {
            this.status = status;
        }

        public void SetDisk(string disk)
        {
            try
            {
                LockDisk();

                if (fsDisk != null) fsDisk.Close();
                fsDisk = null;
                dskchg = true;

                pnlDisk.BackColor = Color.Black;
                txtDisk.Text = "";

                if (disk != null)
                {
                    try
                    {
                        fsDisk = new FileStream(disk, FileMode.Open, FileAccess.ReadWrite, FileShare.None);
                    }
                    catch (UnauthorizedAccessException)
                    {
                        try
                        {
                            fsDisk = new FileStream(disk, FileMode.Open, FileAccess.Read, FileShare.None);
                        }
                        catch
                        {
                            MessageBox.Show(this, "Unable to open disk " + Path.GetFileName(disk), "Error", MessageBoxButtons.OK, MessageBoxIcon.Error);
                            return;
                        }
                    }
                    catch
                    {
                        MessageBox.Show(this, "Unable to open disk " + Path.GetFileName(disk), "Error", MessageBoxButtons.OK, MessageBoxIcon.Error);
                        return;
                    }

                    pnlDisk.BackColor = Color.Blue;
                    txtDisk.Text = Path.GetFileName(disk);
                }
            }
            finally
            {

                UnlockDisk();
            }
        }

        public bool IsReady()
        {
            return fsDisk != null;
        }

        public bool IsWriteProtected()
        {
            return !fsDisk.CanWrite;
        }

        public bool IsDiskChanged()
        {
            bool temp = dskchg;
            dskchg = false;
            return temp;
        }

        public void Seek(int sector)
        {
            fsDisk.Seek(sector * 512, SeekOrigin.Begin);
        }

        public void Truncate()
        {
            fsDisk.SetLength(fsDisk.Position);
        }

        public void Read(byte[] sector)
        {
            fsDisk.Read(sector, 0, 512);
        }

        public void Write(byte[] sector)
        {
            fsDisk.Write(sector, 0, 512);
        }

        private void btnEject_Click(object sender, EventArgs e)
        {
            SetDisk(null);

            ofdDisk.FileName = "";
            if (ofdDisk.ShowDialog(this) == DialogResult.OK)
            {
                SetDisk(ofdDisk.FileName);
            }
        }

        delegate void SetLEDColorCallback(Color color);

        void SetLEDColor(Color color)
        {
            if (this.panelLED.InvokeRequired)
            {
                SetLEDColorCallback d = new SetLEDColorCallback(SetLEDColor);
                this.Invoke(d, new object[] { color });
            }
            else
            {
                switch (status)
                {
                    case 0:
                        panelLED.BackColor = color;
                        break;
                    case -1:
                        panelLED.BackColor = color;
                        break;
                    case 1:
                        panelLED.BackColor = color;
                        break;
                }
            }
        }

        private void tmrStatus_Tick(object sender, EventArgs e)
        {
            SetLEDColor(status == 0 ? Color.DarkGreen :
                        status == 1 ? Color.LightGreen :
                        status == -1 ? Color.Pink : Color.DarkGreen);

        }
    }
}
