using System;
using System.IO;
using System.Collections;
using System.Text;

namespace msxdisk
{
	/// <summary>
	/// Summary description for File.
	/// </summary>
	public class FileHelper
	{
		//Reads the file at [filePath] and returns the contents.
		//Returns null if an error occurs.
		public static string ReadFile(string filePath)
		{
			StreamReader sr = null;
			string       file;

			try 
			{
				sr = new StreamReader(filePath, Encoding.UTF8);
				file = sr.ReadToEnd();
			}
			catch (Exception ex)
			{
				throw ex;
				//file = null;
			}
			finally
			{
				if (sr != null)
					sr.Close();
			}
			return file;
		}

		//Writes [file] to the file at [filePath].
		//Returns true if successful otherwise false
		public static void   WriteFile(string filePath, string file)
		{
			StreamWriter sw = null;
			//bool         success = true;

			try 
			{
				sw = new StreamWriter(filePath, false, Encoding.UTF8);
				sw.Write(file);
			}
			catch (Exception ex)
			{
				throw ex;
				//success = false;
			}
			finally
			{
				if (sw != null)
					sw.Close();
			}
			//return success;
		}

		//Reads a csv value [file] starting from position [pos].
		//This function also accepts value that are enclosed in ".
		//If value is enclosed in ", "" within the value indicates a single ".
		//pos is set to the next value.
		//eof is set to true when end of file is reached otherwise false.
		//eol is set to true when end of line is reached otherwise false.
		public static string GetCsvValue(string file, ref int pos, out bool eof, out bool eol)
		{
            StringBuilder val = new StringBuilder(""); // stores the current value
			char   chr;          // stores the current character
			bool   frst = true;  // indicates that current character is first character
			bool   del  = false; // indicates that value is delimited with "
			bool   dcl  = false; // indicates that last character of delimited value was "
			bool   crlf = false; // indicates that previous character was a cr or a lf
			bool   eov  = false; // indicates end of value

			eof = false;
			eol = false;

			for(;!eov;pos++)
			{
				// if posistion is at end of file
				if (pos == file.Length)
				{
					eof = true;
					eol = true;
					eov = true;
					break;
				}

				// store the current character
				chr = file[pos];

				// if first character is "
				if (frst)
				{
					frst = false;
					if (chr == '"')
					{
						del = true;
						continue;
					}
				}

				// if field is not delimited or last character was "
				if (!del || dcl)
				{
					// if previous character was CR or LF
					if (crlf)
					{
						// if this character is not CR or LF
						if ((chr != (char) 13) && (chr != (char) 10))
						{
							// indicate EOL and EOV and return value
							eol = true;
							eov = true;
							break;
						}
					}

					// if character is CR or LF
					if ((chr == (char) 13) || (chr == (char) 10))
					{
						crlf = true;
						continue;
					}

					// if character is comma
					if (chr == ',')
					{
						eov = true;
						continue;
					}

					// if text is delimited and closed and this point is reached
					if (del)
					{
						// if chr is " ie. double delimiter
						if (chr == '"')
						{
							dcl = false;
							val.Append('"');
							continue;
						}

						// this should be an error because either " or , or CR or LF should follow " in del. value
						// but we are going to handle last " as a normal character and add this character
						dcl = false;
                        val.Append("\"");
                        val.Append(chr);
						continue;
					}
				}

				// if text is delimited and character is "
				if (del && (chr == '"'))
				{
					dcl = true;
					continue;
				}

				// append character
				val.Append(chr);
			}

			return val.ToString();
		}

		//Similar to GetCsvValue
		//This function reads the whole line and returns the values in an string array
		public static string[] GetCsvValues(string file, ref int pos, out bool eof)
		{
			ArrayList temp = new ArrayList();

			eof = false;

			bool eol = false;
			while(!eol)
			{
				string val = GetCsvValue(file, ref pos, out eof, out eol);
				temp.Add(val);
			}

			string[] ret = new string[temp.Count];
			temp.CopyTo(ret);
			return ret;
		}

		//Converts [val] to a csv value enclosed with "".
		public static string ToCsvValue(object val)
		{
            string str;
            if (val is DateTime)
                str = string.Format("{0:yyyy-MM-dd}", val);
            else if (val is Enum)
                str = Convert.ToInt32(val).ToString();
            else if (val == null)
                str = "";
            else
                str = val.ToString();

			StringBuilder ret = new StringBuilder("", Convert.ToInt32(str.Length * 1.1));
            ret.Append("\"");
            for (int i = 0; i < str.Length; i++)
			{
				char chr = str[i];
				if (chr == '"')
					ret.Append("\"\"");
				else
					ret.Append(chr);
			}
			ret.Append("\"");
			return ret.ToString();
		}
	}
}
