﻿using System;
using System.ComponentModel;
using System.Windows.Forms;
using System.IO;
using System.IO.Ports;

using System.Threading;

namespace msxdisk
{
    public partial class MainForm : Form
    {
        const int sleep = 1;

        SerialPort rs232;

        byte[] syncphrase = new byte[] { 0x7A, 0xBB, 0x82, 0xDB };
        byte[] buffer = new byte[14];
        byte[] input = new byte[14];

        ChestForm frmChest = new ChestForm();

        Form activeForm = null;

        public MainForm()
        {
            InitializeComponent();

            DiskDriveA.txtDriveLetter.Text = "A:";
            DiskDriveB.txtDriveLetter.Text = "B:";
        }

        private void recvInput(out byte x, out byte y, out byte a, out byte f, out byte b, out byte c, out byte d, out byte e, out byte h, out byte l)
        {
            int j = 0; //position in input/syncphrase
            while (j < input.Length)
            {
                int bytes;
                try
                {
                    bytes = rs232.Read(buffer, 0, input.Length - j);

                    for (int i = 0; i < bytes; i++)
                    {
                        if (j < syncphrase.Length && syncphrase[j] != buffer[i])
                        {
                            i -= j;
                            j = 0;
                        }
                        else
                        {
                            input[j] = buffer[i];
                            j++;
                        }
                    }
                }
                catch
                {
                    Thread.Sleep(20);
                }
            }

            x = input[5];
            y = input[4];
            a = input[7];
            f = input[6];
            b = input[9];
            c = input[8];
            d = input[11];
            e = input[10];
            l = input[13];
            h = input[12];
        }

        private void MainForm_Shown(object sender, EventArgs e)
        {
            PortForm portForm = new PortForm();
            for (; ;)
            {
                if (portForm.ShowDialog(this) == DialogResult.OK)
                {
                    try
                    {
                        rs232 = new SerialPort(portForm.cmbPort.Text, 57600);
                        rs232.Open();
                        break;
                    }
                    catch
                    {
                        MessageBox.Show(this, "Unable to open port " + portForm.cmbPort.Text, "Error", MessageBoxButtons.OK, MessageBoxIcon.Error);
                    }
                }
                else
                {
                    this.Close();
                    return;
                }
            }

            worker.RunWorkerAsync();
        }

        private void worker_DoWork(object sender, DoWorkEventArgs ev)
        {
            byte[] output = new byte[4];
            byte[] sector = new byte[512];
            byte[] frame = new byte[64];
            int offset;

            for (; ; )
            {
                //Receive Input
                recvInput(out byte x, out byte y, out byte a, out byte f, out byte b, out byte c, out byte d, out byte e, out byte h, out byte l);

                string datetime = string.Format("{0:yyyy-MM-dd HH:mm:ss.fff}", DateTime.Now);
                string device = "";
                string command = "";
                string drive = "";
                string operation = "";
                string sectors = "";
                string sectornumber = "";

                switch (x)
                {
                    case 0x01:
                        device = "DISK";
                        switch (y)
                        {
                            case 0x00:
                                command = "DRIVES?";
                                break;
                            case 0x01:
                                command = "DSKIO";
                                drive = (char)(65 + a) + ":";
                                operation = (f & 0b1) == 0 ? "READ" : "WRITE";
                                sectors = string.Format("{0:X2}", b);
                                sectornumber = string.Format("{0:X2}{1:X2}", d, e);
                                break;
                            case 0x02:
                                command = "DSKCHG?";
                                drive = (char)(65 + a) + ":";
                                break;
                            case 0x03:
                                command = "DSKFMT";
                                drive = (char)(65 + d) + ":";
                                break;
                        }
                        break;

                    case 0x02:
                        device = "TAPE";
                        switch (y)
                        {
                            case 0x00:
                                command = "BIOS";
                                break;
                            case 0x01:
                                command = "TAPION";
                                break;
                            case 0x02:
                                command = "TAPIN";
                                break;
                            case 0x03:
                                command = "TAPIOF";
                                break;
                            case 0x04:
                                command = "TAPOON";
                                break;
                            case 0x05:
                                command = "TAPOUT";
                                break;
                            case 0x06:
                                command = "TAPOOF";
                                break;
                        }
                        break;

                    case 0x03:
                        device = "ROM";
                        command = "";
                        break;

                    case 0x04:
                        device = "PSGOUT";
                        command = "";
                        break;

                    default:
                        device = string.Format("{0:X2}", x);
                        command = string.Format("{0:X2}", y);
                        break;
                }

                Log(new string[]
                {
                    datetime,
                    device,
                    command,
                    drive,
                    operation,
                    sectors,
                    sectornumber,
                });

                switch (x)
                {
                    case 0x01: //disk

                        switch (y)
                        {
                            case 0x00: //drives

                                //Send Output
                                output[1] = 0; //H 
                                output[0] = 2; //L

                                System.Threading.Thread.Sleep(sleep);
                                rs232.Write(output, 0, 2);

                                break;

                            case 0x01: //dskio

                                DiskDrive diskDrive = null;
                                switch (a)
                                {
                                    case 0:
                                        diskDrive = DiskDriveA;
                                        break;
                                    case 1:
                                        diskDrive = DiskDriveB;
                                        break;
                                }

                                try
                                {
                                    diskDrive.LockDisk();

                                    if (!diskDrive.IsReady())
                                    {
                                        //Send Output
                                        output[1] = 2; //A not ready
                                        output[0] = 1; //F error
                                        output[3] = b; //B sectors remaining
                                        output[2] = 0; //C

                                        System.Threading.Thread.Sleep(sleep);
                                        rs232.Write(output, 0, 4);
                                    }
                                    else if ((f & 0b1) != 0 && diskDrive.IsWriteProtected())
                                    {
                                        //Send Output
                                        output[1] = 0; //A write protected
                                        output[0] = 1; //F error
                                        output[3] = b; //B sectors remaining
                                        output[2] = 0; //C

                                        System.Threading.Thread.Sleep(sleep);
                                        rs232.Write(output, 0, 4);
                                    }
                                    else
                                    {
                                        //Send Output
                                        output[1] = 0; //A 
                                        output[0] = 0; //F
                                        output[3] = 0; //B
                                        output[2] = 0; //C

                                        System.Threading.Thread.Sleep(sleep);
                                        rs232.Write(output, 0, 4);

                                        diskDrive.Seek(d << 8 | e);

                                        if ((f & 0b1) == 0)
                                        {
                                            //Read
                                            diskDrive.SetStatus(1);
                                            System.Threading.Thread.Sleep(sleep);
                                            for (int i = 0; i < b; i++)
                                            {
                                                diskDrive.Read(sector);
                                                rs232.Write(sector, 0, 512);
                                            }
                                            diskDrive.SetStatus(0);
                                        }
                                        else
                                        {
                                            //Write
                                            diskDrive.SetStatus(-1);
                                            for (int i = 0; i < b; i++)
                                            {
                                                offset = 0;
                                                while (offset < 512)
                                                    offset += rs232.Read(sector, offset, 512 - offset);

                                                diskDrive.Write(sector);
                                            }
                                            diskDrive.SetStatus(0);
                                        }

                                    }
                                }
                                finally
                                {
                                    diskDrive.UnlockDisk();
                                }

                                break;

                            case 0x02: //dskchg

                                diskDrive = null;
                                switch (a)
                                {
                                    case 0:
                                        diskDrive = DiskDriveA;
                                        break;
                                    case 1:
                                        diskDrive = DiskDriveB;
                                        break;
                                }

                                try
                                {
                                    diskDrive.LockDisk();

                                    if (!diskDrive.IsReady())
                                    {
                                        //Send Output
                                        output[1] = 2; //A not ready
                                        output[0] = 1; //F error
                                        output[3] = 0; //B
                                        output[2] = 0; //C

                                        System.Threading.Thread.Sleep(sleep);
                                        rs232.Write(output, 0, 4);
                                    }
                                    else
                                    {
                                        //Send Output
                                        output[1] = 0; //A 
                                        output[0] = 0; //F
                                        output[3] = diskDrive.IsDiskChanged() ? (byte)0xFF : (byte)0x01; //B
                                        output[2] = 0; //C

                                        System.Threading.Thread.Sleep(sleep);
                                        rs232.Write(output, 0, 4);
                                    }
                                }
                                finally
                                {
                                    diskDrive.UnlockDisk();
                                }

                                break;

                            case 0x03: //dskfmt

                                diskDrive = null;
                                switch (d)
                                {
                                    case 0:
                                        diskDrive = DiskDriveA;
                                        break;
                                    case 1:
                                        diskDrive = DiskDriveB;
                                        break;
                                }

                                try
                                {
                                    diskDrive.LockDisk();

                                    if (!diskDrive.IsReady())
                                    {
                                        //Send Output
                                        output[1] = 2; //A not ready
                                        output[0] = 1; //F error
                                        output[3] = 0; //B
                                        output[2] = 0; //C

                                        System.Threading.Thread.Sleep(sleep);
                                        rs232.Write(output, 0, 4);
                                    }
                                    else if (diskDrive.IsWriteProtected())
                                    {
                                        //Send Output
                                        output[1] = 0; //A write protected
                                        output[0] = 1; //F error
                                        output[3] = 0; //B
                                        output[2] = 0; //C

                                        System.Threading.Thread.Sleep(sleep);
                                        rs232.Write(output, 0, 4);
                                    }
                                    else
                                    {
                                        try
                                        {
                                            FileStream fsBlank = new FileStream("BLANK.DSK", FileMode.Open, FileAccess.Read);
                                            diskDrive.Seek(0);
                                            diskDrive.Truncate();
                                            for (int i = 0; i < fsBlank.Length; i += 512)
                                            {
                                                fsBlank.Read(sector, 0, 512);
                                                diskDrive.Write(sector);
                                            }
                                            fsBlank.Close();

                                            //Send Output
                                            output[1] = 0; //A 
                                            output[0] = 0; //F
                                            output[3] = 0; //B
                                            output[2] = 0; //C

                                            System.Threading.Thread.Sleep(sleep);
                                            rs232.Write(output, 0, 4);
                                        }
                                        catch
                                        {
                                            //Send Output
                                            output[1] = 2; //A not ready
                                            output[0] = 1; //F error
                                            output[3] = 0; //B
                                            output[2] = 0; //C

                                            System.Threading.Thread.Sleep(sleep);
                                            rs232.Write(output, 0, 4);
                                        }
                                    }
                                }
                                finally
                                {
                                    diskDrive.UnlockDisk();
                                }

                                break;
                        }

                        break;

                    case 0x02: //TAPE

                        switch (y)
                        {
                            case 0x00: //BIOS

                                SetEnabled(false);

                                byte[] bios = new byte[32768];

                                offset = 0;
                                while (offset < 32768)
                                    offset += rs232.Read(bios, offset, 32768 - offset);

                                bios[0xE2] = 0xDD;
                                bios[0xE3] = 0x19;

                                bios[0xE5] = 0xEC;
                                bios[0xE6] = 0x19;

                                bios[0xE8] = 0x1F;
                                bios[0xE9] = 0x1A;

                                bios[0xEB] = 0x2B;
                                bios[0xEC] = 0x1A;

                                bios[0xEE] = 0x3A;
                                bios[0xEF] = 0x1A;

                                bios[0xF1] = 0x80;
                                bios[0xF2] = 0x1A;

                                msxdisk.Properties.Resources.TAPEIO.CopyTo(bios, 0x19DD);

                                System.Threading.Thread.Sleep(sleep);
                                rs232.Write(bios, 0, 32768);

                                SetEnabled(true);

                                break;

                            case 0x01: //TAPION

                                if (TapeDrive.TAPION())
                                {
                                    output[1] = 0; //A
                                    output[0] = 0; //F

                                    TapeDrive.SetStatus(1);
                                }
                                else
                                {
                                    output[1] = 0; //A
                                    output[0] = 1; //F
                                }

                                System.Threading.Thread.Sleep(sleep);
                                rs232.Write(output, 0, 2);

                                break;

                            case 0x02: //TAPIN

                                if (TapeDrive.TAPIN(frame))
                                {
                                    output[1] = 0; //A
                                    output[0] = 0; //F

                                    System.Threading.Thread.Sleep(sleep);
                                    rs232.Write(output, 0, 2);

                                    System.Threading.Thread.Sleep(sleep);
                                    rs232.Write(frame, 0, Cassette.frameSize);
                                }
                                else
                                {
                                    output[1] = 0; //A
                                    output[0] = 1; //F

                                    System.Threading.Thread.Sleep(sleep);
                                    rs232.Write(output, 0, 2);
                                }

                                break;

                            case 0x03: //TAPIOF

                                if (TapeDrive.TAPIOF())
                                {
                                    output[1] = 0; //A
                                    output[0] = 0; //F

                                    System.Threading.Thread.Sleep(sleep);
                                    rs232.Write(output, 0, 2);
                                }
                                else
                                {
                                    output[1] = 0; //A
                                    output[0] = 1; //F

                                    System.Threading.Thread.Sleep(sleep);
                                    rs232.Write(output, 0, 2);
                                }

                                TapeDrive.SetStatus(0);

                                break;

                            case 0x04: //TAPOON

                                if (TapeDrive.TAPOON())
                                {
                                    output[1] = a; //A
                                    output[0] = 0; //F

                                    TapeDrive.SetStatus(-1);
                                }
                                else
                                {
                                    output[1] = a; //A
                                    output[0] = 1; //F
                                }

                                System.Threading.Thread.Sleep(sleep);
                                rs232.Write(output, 0, 2);

                                break;

                            case 0x05: //TAPOUT

                                if (TapeDrive.TAPOUT_OK())
                                {
                                    output[1] = 0; //A
                                    output[0] = 0; //F

                                    System.Threading.Thread.Sleep(sleep);
                                    rs232.Write(output, 0, 2);

                                    offset = 0;
                                    while (offset < Cassette.frameSize)
                                        offset += rs232.Read(frame, offset, Cassette.frameSize - offset);

                                    TapeDrive.TAPOUT(frame);
                                }
                                else
                                {
                                    output[1] = 0; //A
                                    output[0] = 1; //F

                                    System.Threading.Thread.Sleep(sleep);
                                    rs232.Write(output, 0, 2);
                                }

                                break;

                            case 0x06: //TAPOOF

                                if (TapeDrive.TAPOOF())
                                {
                                    output[1] = 0; //A
                                    output[0] = 0; //F

                                    System.Threading.Thread.Sleep(sleep);
                                    rs232.Write(output, 0, 2);
                                }
                                else
                                {
                                    output[1] = 0; //A
                                    output[0] = 1; //F

                                    System.Threading.Thread.Sleep(sleep);
                                    rs232.Write(output, 0, 2);
                                }

                                TapeDrive.SetStatus(0);

                                break;
                        }

                        break;

                    case 0x03: //ROM

                        CloseDialog();
                        SendROM();
                        break;

                    case 0x04: //PSGOUT

                        CloseDialog();

                        PSGOUTForm PSGOUTForm = new PSGOUTForm();
                        PSGOUTForm.rs232 = rs232;

                        System.Threading.Thread.Sleep(sleep);
                        ShowDialog(PSGOUTForm);
                        break;

                    case 0xFF:

                        switch (y)
                        {
                            case 0x00: //DMPROM

                                byte[] rom = new byte[32768];

                                offset = 0;
                                while (offset < 32768)
                                    offset += rs232.Read(rom, offset, 32768 - offset);

                                FileStream fsROM = new FileStream("BIOS.ROM", FileMode.Create, FileAccess.ReadWrite);
                                fsROM.Write(rom, 0, 32768);
                                fsROM.Close();

                                break;
                        }

                        break;
                }
                
            }
        }

        private void btnChest_Click(object sender, EventArgs e)
        {
            UseWaitCursor = true;
            activeForm = frmChest;
            if (frmChest.ShowDialog(this) == DialogResult.OK)
            {
                ChestForm.Item item = (ChestForm.Item)frmChest.SelectedItem;

                switch (item.type)
                {
                    case "CAS":

                        TapeDrive.SetCAS(Path.Combine("CHEST", item.file));

                        break;

                    case "DSK":

                        DiskDriveA.SetDisk(Path.Combine("CHEST", item.file));

                        break;

                    case "ROM":

                        if (mnuSendROM.Enabled)
                        {
                            SendROMForm sendROMForm = new SendROMForm();
                            sendROMForm.romfile = Path.Combine("CHEST", item.file);
                            sendROMForm.rs232 = rs232;
                            ShowDialog(sendROMForm);
                            if (sendROMForm.Reset)
                                mnuSendROM.Enabled = false;

                        }
                        else
                        {
                            MessageBox.Show(this, "MSX is not waiting for ROM.", "Error", MessageBoxButtons.OK, MessageBoxIcon.Error);
                        }

                        break;
                }
            }
            UseWaitCursor = false;
        }

        private void mnuPlayROMLDR_Click(object sender, EventArgs e)
        {
            UseWaitCursor = true;
            PlayCASForm playCASForm = new PlayCASForm();
            playCASForm.casfile = "ROMLDR.CAS";
            ShowDialog(playCASForm);
            UseWaitCursor = false;
        }

        private void mnuPlayCASLDR_Click(object sender, EventArgs e)
        {
            UseWaitCursor = true;
            PlayCASForm playCASForm = new PlayCASForm();
            playCASForm.casfile = "CASLDR.CAS";
            ShowDialog(playCASForm);
            UseWaitCursor = false;
        }

        private void mnuPlayPSGOUT_Click(object sender, EventArgs e)
        {
            UseWaitCursor = true;
            PlayCASForm playCASForm = new PlayCASForm();
            playCASForm.casfile = "PSGOUT.CAS";
            ShowDialog(playCASForm);
            UseWaitCursor = false;
        }

        private void mnuPlayOtherCAS_Click(object sender, EventArgs e)
        {
            UseWaitCursor = true;
            if (ofdCAS.ShowDialog(this) == DialogResult.OK)
            {
                PlayCASForm playCASForm = new PlayCASForm();
                playCASForm.casfile = ofdCAS.FileName;
                ShowDialog(playCASForm);
            }
            UseWaitCursor = false;
        }

        private void mnuSendDiskROM_Click(object sender, EventArgs e)
        {
            UseWaitCursor = true;
            SendROMForm sendROMForm = new SendROMForm();
            sendROMForm.romfile = "DISK.ROM";
            sendROMForm.rs232 = rs232;
            ShowDialog(sendROMForm);
            if (sendROMForm.Reset)
                mnuSendROM.Enabled = false;
            UseWaitCursor = false;
        }

        private void mnuSendOtherROM_Click(object sender, EventArgs e)
        {
            UseWaitCursor = true;
            if (ofdROM.ShowDialog(this) == DialogResult.OK)
            {
                SendROMForm sendROMForm = new SendROMForm();
                sendROMForm.romfile = ofdROM.FileName;
                sendROMForm.rs232 = rs232;
                ShowDialog(sendROMForm);
                if (sendROMForm.Reset)
                    mnuSendROM.Enabled = false;

            }
            UseWaitCursor = false;
        }

        private void btnLog_Click(object sender, EventArgs e)
        {
            btnLog.Checked = !btnLog.Checked;
            if (btnLog.Checked)
            {
                this.ClientSize = new System.Drawing.Size(469, 556);
            }
            else
            {
                this.ClientSize = new System.Drawing.Size(469, 315);
            }
        }

        delegate void LogCallback(string[] log);

        private void Log(string[] log)
        {
            if (this.lstLog.InvokeRequired)
            {
                LogCallback d = new LogCallback(Log);
                this.Invoke(d, new object[] { log });
            }
            else
            {
                lstLog.BeginUpdate();
                ListViewItem lviLog = new ListViewItem(log);
                this.lstLog.Items.Insert(0, lviLog);
                this.lstLog.EnsureVisible(0);
                if (lstLog.Items.Count > 100)
                    this.lstLog.Items.RemoveAt(100);
                lstLog.EndUpdate();
            }
        }

        delegate void SetEnabledCallback(bool enabled);

        private void SetEnabled(bool enabled)
        {
            if (this.InvokeRequired)
            {
                SetEnabledCallback d = new SetEnabledCallback(SetEnabled);
                this.Invoke(d, new object[] { enabled });
            }
            else
            {
                this.Enabled = enabled;
                this.UseWaitCursor = !enabled;
            }
        }

        delegate DialogResult ShowDialogCallback(Form form);

        private DialogResult ShowDialog(Form form)
        {
            if (this.InvokeRequired)
            {
                ShowDialogCallback d = new ShowDialogCallback(ShowDialog);
                return (DialogResult)this.Invoke(d, new object[] { form });
            }
            else
            {
                activeForm = form;
                DialogResult dialogResult = form.ShowDialog(this);
                activeForm = null;
                return dialogResult;
            }
        }

        delegate void CloseDialogCallback();

        private void CloseDialog()
        {
            if (activeForm != null)
                if (this.InvokeRequired)
                {
                    CloseDialogCallback d = new CloseDialogCallback(CloseDialog);
                    this.Invoke(d, new object[] { });
                }
                else
                {
                    activeForm.Close();
                }
        }

        delegate void SendROMCallback();

        private void SendROM()
        {
            if (this.InvokeRequired)
            {
                SendROMCallback d = new SendROMCallback(SendROM);
                this.Invoke(d, new object[] { });
            }
            else
            {
                mnuSendROM.Enabled = true;
                mnuSendROM.ShowDropDown();
            }
        }
    }
}
