﻿using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using System.Windows.Forms;

using System.Threading;
using System.IO.Ports;

using NAudio.Wave;
using NAudio.Dsp;

namespace msxdisk
{
    public partial class PSGOUTForm : Form
    {
        const float samplesPerSecond = 5760f;
        const int bufferLength = 576;
        const int buffers = 3;
        const int maxBuffers = 6;

        const int speedAutoEvery = 10;
        const int speedAutoDelay = 0;

        public PSGOUTForm()
        {
            InitializeComponent();
        }

        public SerialPort rs232;

        WasapiLoopbackCapture capture = new WasapiLoopbackCapture();
        Queue<byte[]> q = new Queue<byte[]>();
        
        readonly byte[] scale = new byte[] { 0, 1, 2, 3, 5, 7, 10, 15, 22, 31, 44, 63, 90, 127, 180, 255 };
        readonly byte[] inv4 = new byte[] { 0, 8, 4, 12, 2, 10, 6, 14, 1, 9, 5, 13, 3, 11, 7, 15 };
        readonly byte[] inv2 = new byte[] { 0, 2, 1, 3 };

        Thread SendThread;

        float blockSize;
        int curBlockSize;

        private void frmPSGOUT_Load(object sender, EventArgs e)
        {
            prbQLength.Maximum = maxBuffers * speedAutoEvery;

            try
            {
                string[] values = System.IO.File.ReadLines("PSGOUT.TXT").ToArray();

                SetVolume(Convert.ToSingle(values[0]));
                SetSpeed(Convert.ToSingle(values[1]));
                chkSpeedAuto.Checked = Convert.ToBoolean(values[2]);
            }
            catch
            {
                SetVolume(1);
                SetSpeed(1);
                chkSpeedAuto.Checked = true;
            }

            SendThread = new Thread(SendWorker);
            SendThread.Start();

            blockSize = capture.WaveFormat.Channels * capture.WaveFormat.SampleRate / samplesPerSecond * speed;
            curBlockSize = (int)(blockSize / capture.WaveFormat.Channels) * capture.WaveFormat.Channels;

            capture.DataAvailable += capture_DataAvailable;
            capture.StartRecording();
        }

        private void trbVolume_Scroll(object sender, EventArgs e)
        {
            SetVolume(trbVolume.Value * trbVolume.Value / 10000f);
        }

        private void trbSpeed_Scroll(object sender, EventArgs e)
        {
            SetSpeed(trbSpeed.Value / 1000f);
        }

        private void PSGOUTForm_FormClosing(object sender, FormClosingEventArgs e)
        {
            capture.StopRecording();

            SendThread.Abort();
            SendThread.Join();

            rs232.Write(new byte[] { 0b11110000 }, 0, 1);

            try
            {
                System.IO.File.WriteAllLines("PSGOUT.TXT", new string[] {
                    volume.ToString(),
                    speed.ToString(),
                    chkSpeedAuto.Checked.ToString(),
                });
            }
            catch
            {

            }
        }

        float volume = 1.0f;
        float speed = 1.000f;

        float[] ieee = new float[65536];
        byte[] buffer = new byte[bufferLength];
        byte[] buffer_ = new byte[bufferLength];
        float y = 0.0f;
        int i = 0;
        int j = 0;

        private void capture_DataAvailable(object sender, WaveInEventArgs e)
        {
            if (e.BytesRecorded == 0)
            {
                q.Clear();

                buffersPlayed = 0;
                sumSpeed = 0;
                avgQLength = 0;

                return;
            }

            Buffer.BlockCopy(e.Buffer, 0, ieee, 0, e.BytesRecorded);

            int samples = e.BytesRecorded / 4;

            for (int x = 0; x < samples; x++)
            {
                y += ieee[x];
                i++;
                if (i == curBlockSize)
                {
                    y /= curBlockSize;
                    y *= volume;

                    buffer[j] = Float2Byte(y);
                    j++;
                    if (j == buffer.Length)
                    {
                        q.Enqueue(buffer);
                        buffer = new byte[bufferLength];

                        j = 0;
                    }

                    blockSize -= curBlockSize;
                    blockSize += capture.WaveFormat.Channels * capture.WaveFormat.SampleRate / samplesPerSecond * speed;
                    curBlockSize = (int)(blockSize / capture.WaveFormat.Channels) * capture.WaveFormat.Channels;

                    y = 0.0f;
                    i = 0;
                }
            }
        }

        float sumSpeed = 0f;
        int buffersPlayed = 0;

        float avgQLength = 0;
        float avgBuffers = 0;

        byte[] state = new byte[] { 0, 0, 0 };

        void SendWorker()
        {
            for (; ; )
            {
                if (q.Count == 0)
                {
                    buffersPlayed = 0;
                    sumSpeed = 0;
                    avgQLength = 0;

                    Level(0);
                    QLength(avgQLength);

                    while (q.Count < buffers)
                    {
                        Thread.Sleep((int)((buffers - q.Count) * bufferLength * 1000 / samplesPerSecond));
                    }
                }
                else if (q.Count > maxBuffers)
                {
                    buffersPlayed = 0;
                    sumSpeed = 0;
                    avgQLength = 0;

                    while (q.Count > buffers) q.Dequeue();
                }
                else
                {
                    byte[] buffer = q.Dequeue();

                    int level = 0;

                    for (int b = 0; b < bufferLength; b++)
                    {
                        byte z = buffer[b];

                        int abs = Math.Abs(z - 128);
                        if (abs > level)
                            level = abs;

                        int t = 0;
                        for(int c = 0; c < state.Length; c++)
                        {
                            t += scale[state[c]];
                        }

                        int d = z - t;

                        int channel = 0;
                        int volume = state[channel];
                        int diff = d;

                        if (d >= 0) //inc
                        {
                            for (int c = 0; c < state.Length; c++)
                            {
                                int e = state[c]; //V4
                                int f = scale[e]; //V8
                                for (int g = e + 1; g < scale.Length; g++)
                                {
                                    int h = scale[g]; //V8 NEW
                                    int k = h - f; //V8 INC
                                    int m = d - k;
                                    if (m < 0) break;
                                    else if (m < diff)
                                    {
                                        channel = c;
                                        volume = g;
                                        diff = m;
                                    }
                                }
                            }
                        }
                        else //dec d is -ve
                        {
                            for (int c = 0; c < state.Length; c++)
                            {
                                int e = state[c]; //V4
                                int f = scale[e]; //V8
                                for (int g = e - 1; g >= 0; g--)
                                {
                                    int h = scale[g]; //V8 NEW
                                    int k = h - f; //V8 INC -ve
                                    int m = d - k; //-10 - -8 = -2
                                    if (m > 0) break;
                                    else if (m > diff)
                                    {
                                        channel = c;
                                        volume = g;
                                        diff = m;
                                    }
                                }
                            }
                        }

                        buffer_[b] = (byte)(0b11000000 | inv2[channel] << 4 | inv4[volume]);

                        state[channel] = (byte)volume;
                    }

                    Level(level);

                    rs232.Write(buffer_, 0, buffer_.Length);

                    buffersPlayed++;

                    avgQLength += (float)q.Count / speedAutoEvery;

                    if (buffersPlayed == speedAutoEvery)
                    {
                        avgBuffers = avgQLength;
                    }

                    if (buffersPlayed % speedAutoEvery == 0)
                    {
                        sumSpeed += speed;

                        int count = buffersPlayed / speedAutoEvery;
                        float offset = avgQLength - avgBuffers;

                        if (count >= speedAutoDelay)
                        {
                            float avgReadSpeed = sumSpeed / count;
                            float playSpeed = buffersPlayed / (offset + buffersPlayed);
                            float newSpeed = avgReadSpeed / playSpeed;

                            if (chkSpeedAuto.Checked)
                                SetSpeed(newSpeed);
                        }

                        QLength(avgQLength);

                        avgQLength = 0;
                    }
                }
            }
        }

        delegate void SetVolumnCallback(float volume);

        private void SetVolume(float volume)
        {
            if (this.lblVolume.InvokeRequired)
            {
                SetVolumnCallback d = new SetVolumnCallback(SetVolume);
                this.Invoke(d, new object[] { volume });
            }
            else
            {
                this.volume = volume;

                int value = (int)(Math.Sqrt(volume) * 100);
                if (value < trbVolume.Minimum) value = trbVolume.Minimum;
                if (value > trbVolume.Maximum) value = trbVolume.Maximum;
                trbVolume.Value = value;

                lblVolume.Text = String.Format("{0:0}%", volume * 100);
            }
        }

        delegate void SetSpeedCallback(float speed);

        private void SetSpeed(float speed)
        {
            if (this.lblSpeed.InvokeRequired)
            {
                SetSpeedCallback d = new SetSpeedCallback(SetSpeed);
                this.Invoke(d, new object[] { speed });
            }
            else
            {
                this.speed = speed;
                int value = (int)(speed * 1000);
                if (value < trbSpeed.Minimum) value = trbSpeed.Minimum;
                if (value > trbSpeed.Maximum) value = trbSpeed.Maximum;
                trbSpeed.Value = value;
                lblSpeed.Text = String.Format("{0:0.000}X", speed);
            }
        }

        delegate void QLengthCallback(float qLength);

        private void QLength(float qLength)
        {
            if (this.InvokeRequired)
            {
                QLengthCallback d = new QLengthCallback(QLength);
                this.Invoke(d, new object[] { qLength });
            }
            else
            {
                int value = (int)Math.Round(qLength * speedAutoEvery);
                if (value > prbQLength.Maximum) value = prbQLength.Maximum;
                prbQLength.Value = value;
            }
        }

        delegate void LevelCallback(int level);

        private void Level(int level)
        {
            if (this.InvokeRequired)
            {
                LevelCallback d = new LevelCallback(Level);
                this.Invoke(d, new object[] { level });
            }
            else
            {
                int value = level;
                if (value > prbLevel.Maximum) value = prbLevel.Maximum;
                prbLevel.Value = value;
            }
        }
        byte Float2Byte(float y)
        {
            return (byte)Math.Round(Math.Min(Math.Max(y * 127.0F + 128.0, 0.0F), 255.0F));
        }

    }
}
